import jsonschema
import logging
from collections import namedtuple

from .constants import Schema, Options

log = logging.getLogger(__name__)

# nmt = nest_meta_tuple
nmt = namedtuple('nmt', ['action_map', 'lookup_map', 'adj_list'])


class TransformNest:
    """
    This class processes the nest, that can be used
    to build an Anthill
    - I/P Nest:
    ...... {
    ......     "node": [{
    ......         "rule": "rule_1.json"
    ......     }],
    ......     "action_1": [{
    ......         "rule": "rule_2.json"
    ......     }, {
    ......         "depends_on": ["node"]
    ......     }]
    ...... }
    """
    action_map = None
    lookup_map = None
    adj_list = None

    def __init__(self, nest):
        self.nest = nest

    def is_valid(self):
        """
        Check Validity of pre-processed nest configuration
        TODO: Check for cyclic dependency
        TODO: Apart from nest node, no other action can have empty dependency
        TODO: Other validity requirements for the nest
        """
        try:
            jsonschema.validate(
                instance=self.nest,
                schema=Schema.NEST_SCHEMA
            )
        except jsonschema.exceptions.ValidationError as err:
            log.error(f"Invalid nest ... \n {err.message}")
            return False
        return True

    def gen_action_map(self):
        """
        Generate mapping for
            <action_label, index>
        """
        return {
            key:value for key, value in enumerate(list(self.nest.keys()))
        }

    def _dependencies(self, action):
        """
        Returns the count of dependencies for the provided action
        """
        _count = [
            len(x[Options.DEPENDS_ON]) for x in action if Options.DEPENDS_ON in x.keys()
        ]
        if not _count:
            return 0
        return _count[0]

    def _get_action(self, inx):
        """
        Returns action corresponding to the index in
        action map
        """
        action_label = self.action_map.get(inx)
        return self.nest.get(action_label)

    def _get_inx(self, label):
        """
        Returns inx corresponding to action label
        from action_map
        """
        for _inx, _label in self.action_map.items():
            if _label == label:
                return _inx

    def gen_lookup_map(self):
        """
        Lookup map maps index to the total number
        of dependencies
        """
        lookup = {}
        for inx in self.action_map:
            _action = self._get_action(inx)
            lookup[inx] = self._dependencies(_action)
        return lookup

    def _dependents(self, inx):
        """
        Returns list of inx which are dependent on
        the provided inx
        """
        res = []
        action_label = self.action_map.get(inx)
        for _label in self.nest:
            _it = self.nest.get(_label)
            _it_dependencies = [
                x[Options.DEPENDS_ON] for x in _it if Options.DEPENDS_ON in x.keys()
            ]
            if _it_dependencies and action_label in _it_dependencies[0]:
                res.append(self._get_inx(_label))
        return res

    def gen_adj_list(self):
        """
        Generate adjancy list to be used to develop topology
        """
        adj_list = {}
        for inx in self.action_map:
            adj_list[inx] = self._dependents(inx)
        return adj_list

    def generate(self):
        """
        Generate all the required META parameters
        for building the Anthill
        """
        self.action_map = self.gen_action_map()
        self.lookup_map = self.gen_lookup_map()
        self.adj_list = self.gen_adj_list()

        return nmt(
            self.action_map,
            self.lookup_map,
            self.adj_list
        )

    def transform(self):
        """
        Transform the nest into engine understandable blueprint
        for building the anthill
        """
        if not self.is_valid():
            return
        return self.generate()





class Schema:

    NEST_SCHEMA = {
        "type": "object",
        "properties": {
            "node": {
                "type": "array",
                "items": [
                    {
                        "type": "object",
                        "properties": {
                            "rule": {
                                "type": "string"
                            }
                        },
                        "required": [
                            "rule"
                        ]
                    }
                ]
            }
        },
        "required": [
            "node"
        ]
    }


class Options:

    DEPENDS_ON = "depends_on"



from anthill.src.topology.topology import Anthill


class Nest:
    """
    Class to handle all operations on nest including build and execution
    """
    instance = None

    def __init__(self, nest):
        self.nest = nest

    def build(self):
        self.instance = Anthill(nest=self.nest).build()

    def run(self):
        pass



"""
Core scripts to handle cli instructions
"""

import sys
import click
import yaml
import json

from .management import Nest
from anthill import __version__


@click.group()
def main():
    """
    Orchastrate. Build. Run.
    """
    pass


@main.command()
def version():
    click.echo("Version: " + __version__)

@main.command()
@click.argument('nest', type=click.File('r'))
@click.option('--run', '-r', is_flag=True, help="Run the anthill after build")
def build(nest, run):
    """
    Build anthill using the yml file provided
    """
    _nest = json.dumps(yaml.safe_load(nest))
    nest_obj  = Nest(nest=_nest)
    try:
        nest_obj.build()
    except Exception as err:
        click.echo(f"Aborting build ... \n {err.message}")
    else:
        click.echo("Anthill built successfully !")
        if run:
            try:
                nest_obj.run()
            except Exception as err:
                click.echo(f"Aborting execution ... \n {err.message}")
            else:
                click.echo("Anthill initiated successfully !")
    return 0



import json
import logging

from .transform import TransformNest

log = logging.getLogger(__name__)


class Anthill:
    """
    Anthill is a topoLOGICAL order of ants.
    Ants are the basic block, consisting of singular functions
    These singular functions whene arranges in a logical order form an
    anthill.
    - Anthill:
    [
        {
            "label": "",
            "meta": {
                "rule": "",
                "depends_on": []
            }
        }
    ]
    """
    _nest = None
    anthill = []
    _anthill_inx = []
    action_map = None
    lookup_map = None
    adj_list = None

    def __init__(self, nest):
        self.nest = json.loads(nest)

    def assign_meta_variables(self, _nmt):
        self.action_map = _nmt.action_map
        self.lookup_map = _nmt.lookup_map
        self.adj_list = _nmt.adj_list

    def get_anthill_action(self, inx):
        """
        Get action agaist inx
        """
        label = self.action_map[inx]
        _meta = {}
        for item in self.nest[label]:
            _meta.update(**item)
        return dict(
            label=label,
            meta=_meta
        )

    def set_next_inx(self):
        """
        Set next action in topological order
        """
        for inx in self.lookup_map:
            if inx not in self._anthill_inx and self.lookup_map[inx]==0:
                self._anthill_inx.append(inx)
                return inx
        return -1

    def update_lookup_map(self, inx):
        """
        Updates incoming branch count from lookup map
        """
        _dependents = self.adj_list[inx]
        for it in _dependents:
            self.lookup_map[it] -= 1

    def _build(self):
        """
        Build anthill from _anthill_inx
        """
        self.anthill = [self.get_anthill_action(x) for x in self._anthill_inx]

    def build(self):
        self.assign_meta_variables(
            TransformNest(nest=self.nest).transform()
        )
        while True:
            inx  = self.set_next_inx()
            if inx==-1:
                break
            self.update_lookup_map(inx)
        self._build()
        return self.anthill
